//
// SizeableDlg
//
// Written by Laurent Kemp (lkempe@netcourrier.com) for Tech Head WebSite
// Copyright (c) 2001 Laurent Kemp.
//
// This code may be used in compiled form in any way you desire. This
// file may be redistributed by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name is included. 
//
// This file is provided "as is" with no expressed or implied warranty.
// The author accepts no liability if it causes any damage to you or your
// computer whatsoever. It's free, so don't hassle me about it.
//
// Origin: July 2001 issue of MSDN Magazine
//         Paul DiLascia 
//         Windows UI: Our WinMgr Sample Makes Custom Window Sizing Simple
//
///////////////////////////////////////////////////////////////////////////////
// Generic sizeable dialog. To create a sizeable dialog, derive from this and
// give your window map to constructor. Otherwise (if you don't derive from
// CSizeableDlg), mimic OnInitDialog, OnSize, OnGetMinMaxInfo.
//

#include "WinMgr.h"

template< class T = CWindow >
class CDialogSizeable
{
public:
	
	CDialogSizeable( WINRECT* pWinMap ) : 
		m_winMgr(pWinMap)
	{
	};

	~CDialogSizeable()
	{
	};

	//////////////////////////////////////////////////////////////////////////
	
	BEGIN_MSG_MAP_EX(CDialogSizeable)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_SIZE(OnSize)
		MSG_WM_GETMINMAXINFO(OnGetMinMaxInfo)
		MSG_WM_DESTROY(OnDestroy)
	END_MSG_MAP()

	//////////////////////////////////////////////////////////////////////////

	LRESULT	OnInitDialog( HWND hwnd, LPARAM lParam )
	{
		BOOL bRet = TRUE;

		T* pT = static_cast<T*>(this);

		m_winMgr.InitToFitSizeFromCurrent(pT);	// make tofit = current size
		m_winMgr.CalcLayout(pT);				// recalc
		m_winMgr.SetWindowPositions(pT);		// set positions

		// Get initial dialog client size
		RECT rectDlg;
		pT->GetClientRect(&rectDlg);

		// shouldn't exist already
		ATLASSERT(!::IsWindow(pT->GetDlgItem(ATL_IDW_STATUS_BAR)));
		if(!::IsWindow(pT->GetDlgItem(ATL_IDW_STATUS_BAR)))
		{
			CWindow wndGripper;
			wndGripper.Create( _T("SCROLLBAR"), 
							   pT->m_hWnd, 
							   rectDlg, 
							   NULL, 
							   WS_CHILD | 
							   WS_VISIBLE | 
							   WS_CLIPSIBLINGS | 
							   SBS_SIZEBOX | 
							   SBS_SIZEGRIP | 
							   SBS_SIZEBOXBOTTOMRIGHTALIGN, 
							   0, 
							   ATL_IDW_STATUS_BAR );

			ATLASSERT(wndGripper.IsWindow());
		}

		return bRet;
	}

	void OnDestroy()
	{
		T* pT = static_cast<T*>(this);

		CWindow wndGripper = pT->GetDlgItem(ATL_IDW_STATUS_BAR);

		if (wndGripper.IsWindow()) 
			wndGripper.DestroyWindow();
	}

	void OnSize(UINT nType, CSize size)
	{
		T* pT = static_cast<T*>(this);

		m_winMgr.CalcLayout(size.cx,size.cy,pT);
		m_winMgr.SetWindowPositions(pT);

		CWindow wndGripper = pT->GetDlgItem(ATL_IDW_STATUS_BAR);

		RECT	rc;
		int		l = ::GetSystemMetrics(SM_CXHSCROLL);
		int		h = ::GetSystemMetrics(SM_CYHSCROLL);

		ATLASSERT(pT->m_hWnd != 0);
		::GetClientRect(pT->m_hWnd, &rc);
		rc.left = rc.right - l;
		rc.top = rc.bottom - h;

		wndGripper.MoveWindow( &rc );

		if(nType == SIZE_MAXIMIZED)
			wndGripper.ShowWindow(SW_HIDE);
		else if(nType == SIZE_RESTORED)
			wndGripper.ShowWindow(SW_SHOW);
	}

	void OnGetMinMaxInfo(LPMINMAXINFO lpMMI)
	{
		T* pT = static_cast<T*>(this);

		m_winMgr.GetMinMaxInfo(pT, lpMMI);
	}

protected:

	CWinMgr m_winMgr;	  // window manager
};
